#!/usr/bin/env python

# NOTE: for debugging of this setup.py script
# set the DISTUTILS_DEBUG environment variable to TRUE
# (or any other non-empty string)

"""
This setup script can be used to build and install the pybufr-ecmwf
module on your system.
"""

#  #[ imported modules
from __future__ import (absolute_import, division,
                        print_function) # , unicode_literals)

import os
import sys
from distutils.core import setup, Extension
from distutils import log
from distutils.errors import DistutilsSetupError
# see: /usr/lib64/python2.6/distutils/errors.py
# for all available error classes

# import build and build_ext using a different name,
# to allow subclassing them
from distutils.command.build import build as _build
from distutils.command.build_ext import build_ext as _build_ext
#from distutils.command.install import install as _install
from distutils.command.install_lib import install_lib as _install_lib

# patch distutils if it can't cope with the "classifiers" or
# "download_url" keywords
from sys import version
if version < '2.2.3':
    from distutils.dist import DistributionMetadata
    DistributionMetadata.classifiers = None
    DistributionMetadata.download_url = None
#  #]

# an alternative might be to use a setup version that seems present
# in the numpy module, see:
# http://www2-pcmdi.llnl.gov/cdat/tutorials/f2py-wrapping-fortran-code/\
#   part-4-packaging-all-this-into-something-that-can-be-distributed-\
#   very-advanced
# and http://www.scipy.org/Documentation/numpy_distutils
# the import would then be
# from numpy.distutils.core import setup, Extension

# suppres some pylint warnings caused by the internal implementation
# of the build class
#pylint: disable-msg=W0201,R0902,R0904

# note that pylint still complains about not being able to load the
# build_interface module, even though the setup routine itself
# seems to handle it just fine.

# modify the build class to allow some custom commandline
# and setup.cfg options to the build stage
class Build(_build):
    #  #[ custom build
    """Adapted Python binary builder."""
    user_options = _build.user_options
    user_options.append(("preferred-fortran-compiler=", None,
                         "name of preferred fortran compiler to be used"))
    user_options.append(("preferred-c-compiler=", None,
                         "name of preferred c compiler to be used"))
    user_options.append(("fortran-compiler=", None,
                         "name and full path of fortran compiler to be used"))
    user_options.append(("fortran-ld-library-path=", None,
                         "path in which shared objects can be found that "+\
                         "are needed by the choosen fortran compiler"))
    user_options.append(("fortran-flags=", None,
                         "flags to be passed to the fortran compiler"))
    user_options.append(("c-compiler=", None,
                         "name and full path of c compiler to be used"))
    user_options.append(("c-ld-library-path=", None,
                         "path in which shared objects can be found that "+\
                         "are needed by the choosen c compiler"))
    user_options.append(("c-flags=", None,
                         "flags to be passed to the c compiler"))
    user_options.append(("download-library-sources=", None,
                         "try to download the latest sources of "+\
                         "the ECMWF BUFR library or not"))

    def initialize_options(self):
        #  #[ initialise the additional options
        """ initialise custom defined options """
        self.preferred_fortran_compiler = None
        self.preferred_c_compiler = None
        self.fortran_compiler = None
        self.fortran_ld_library_path = None
        self.fortran_flags = None
        self.c_compiler = None
        self.c_ld_library_path = None
        self.c_flags = None
        self.download_library_sources = None

        _build.initialize_options(self)
        #  #]
    def run(self):
        #  #[ modified run (for debugging only)
        # some test prints:
        print("build: self.user_options = ", self.user_options)
        print("build: self.preferred_fortran_compiler = ",
              self.preferred_fortran_compiler)
        print("build: self.preferred_c_compiler = ",
              self.preferred_c_compiler)
        print('build: self.download_library_sources = ',
              self.download_library_sources)

        # call the run command of the default build
        _build.run(self)
        #  #]
    #  #]

# modify the build_ext class to allow some custom commandline
# and setup.cfg options to the build_ext stage
class BuildExt(_build_ext):
    #  #[ custom build_ext

    """Specialized Python extension builder."""
    # implement whatever needs to be different...
    # see the original build_ext.py in:
    #    /usr/lib64/python2.6/distutils/command/build_ext.py
    # see also the instructions in:
    #    http://docs.python.org/distutils/extending.html

    # todo: see if I can modify any of the compiler classes
    # like /usr/lib64/python2.6/distutils/*compiler.py
    # to handle fortran as well

    user_options = _build_ext.user_options
    user_options.append(("preferred-fortran-compiler=", None,
                         "name of fortran compiler to be used"))
    user_options.append(("preferred-c-compiler=", None,
                         "name of c compiler to be used"))
    user_options.append(("fortran-compiler=", None,
                         "name and full path of fortran compiler to be used"))
    user_options.append(("fortran-ld-library-path=", None,
                         "path in which shared objects can be found that "+\
                         "are needed by the choosen fortran compiler"))
    user_options.append(("fortran-flags=", None,
                         "flags to be passed to the fortran compiler"))
    user_options.append(("c-compiler=", None,
                         "name and full path of c compiler to be used"))
    user_options.append(("c-ld-library-path=", None,
                         "path in which shared objects can be found that "+\
                         "are needed by the choosen c compiler"))
    user_options.append(("c-flags=", None,
                         "flags to be passed to the c compiler"))
    user_options.append(("download-library-sources=", None,
                         "try to download the latest sources of "+\
                         "the ECMWF BUFR library or not"))

    def initialize_options(self):
        #  #[ initialise the additional options
        """ initialise custom defined options """
        self.preferred_fortran_compiler = None
        self.preferred_c_compiler = None
        self.fortran_compiler = None
        self.fortran_ld_library_path = None
        self.fortran_flags = None
        self.c_compiler = None
        self.c_ld_library_path = None
        self.c_flags = None
        self.download_library_sources = None

        _build_ext.initialize_options(self)
        #  #]
    def finalize_options(self):
        #  #[ make settings available
        """ copy user defined options from the build to the
        build_ext class instance """
        # this copies the user_options from the build
        # to the build_ext class, so I'll have to modify
        # the build class as well to allow new options
        self.set_undefined_options('build',
                                   ('preferred_fortran_compiler',
                                    'preferred_fortran_compiler'),
                                   ('preferred_c_compiler',
                                    'preferred_c_compiler'),
                                   ('fortran_compiler',
                                    'fortran_compiler'),
                                   ('fortran_ld_library_path',
                                    'fortran_ld_library_path'),
                                   ('fortran_flags',
                                    'fortran_flags'),
                                   ('c_compiler',
                                    'c_compiler'),
                                   ('c_ld_library_path',
                                    'c_ld_library_path'),
                                   ('c_flags',
                                    'c_flags'),
                                   ('download_library_sources',
                                    'download_library_sources')
                                   )
        _build_ext.finalize_options(self)
        #  #]
    #def run(self):
        #  #[ modified run (for debugging only)
        # (actually I dont know how to build fortran this way,
        #  so I'll use my own build script here in stead
        #  _build_ext.run(self)  )

        # test prints
        #print("python executable: ", sys.executable)
        #print("build_ext: self.user_options = ", self.user_options)
        #print("build_ext: self.preferred_fortran_compiler = ",
        #      self.preferred_fortran_compiler)
        #print("build_ext: self.preferred_c_compiler = ",
        #      self.preferred_c_compiler)
        #print("build_ext: self.fortran_compiler = ",
        #      self.fortran_compiler)
        #print("build_ext: self.fortran_ld_library_path = ",
        #      self.fortran_ld_library_path)
        #print("build_ext: self.fortran_flags = ",
        #      self.fortran_flags)
        #print("build_ext: self.c_compiler = ",
        #      self.c_compiler)
        #print("build_ext: self.c_ld_library_path = ",
        #      self.c_ld_library_path)
        #print("build_ext: self.c_flags = ",
        #      self.c_flags)
        #print("build_ext: self.download_library_sources = ",
        #      self.download_library_sources)

        # this run command in turn runs the build_extension method
        #_build_ext.run(self)
        #  #]
    def build_extension(self, ext):
        #  #[ the actual build

        """ initiate building the extension module """
        # don't re-initialise these properties here !
        # at his point the setup machinery already has processed
        # the cfg file and the commandline, so these will be lost
        # if you init these properties here again!
        #
        #self.preferred_fortran_compiler = None
        #self.preferred_c_compiler = None
        #self.fortran_compiler = None
        #self.fortran_ld_library_path = None
        #self.fortran_flags = None
        #self.c_compiler = None
        #self.c_ld_library_path = None
        #self.c_flags = None
        #self.download_library_sources = None

        #fullname = self.get_ext_fullname(ext.name)
        #print("trying to build extension: ", fullname)
        log.info("building '%s' extension", ext.name)

        #print("ext.sources = ", ext.sources)

        # this does not work properly yet for setup.py bdist
        # since in that case the pybufr_ecmwf/ecmwfbufr.so
        # needs to be created below build/lib.linux-i686-2.6/
        # so inspect the path settings for the build:

        build_dir = os.path.join(self.build_lib, "pybufr_ecmwf")
        build_dir = os.path.abspath(build_dir)
        #print("self.build_lib = ", self.build_lib)
        print("initiating build in dir: ", build_dir)

        #if os.path.isdir(build_dir):
        #    sys.path.append(build_dir)
        if os.path.isdir(build_dir):
            sys.path.append(build_dir)
        else:
            raise DistutilsSetupError( \
                  "could not find directory in which the module should"
                  "be build. Something seems wrong in setup.py."
                  "Please report this to the developer of this module.")

        # this enters the automatic build system, which is what I don't
        # want at the moment, since it seems not to handle fortran
        #_build_ext.build_extension(self, ext)

        cwd = os.getcwd()
        os.chdir(build_dir)

        # find parent dir that holds build_interface.py and
        # force path to include that dir
        base_build_dir = None
        tmp_dir = build_dir
        while base_build_dir is None:
            files = os.listdir(tmp_dir)
            if 'build_interface.py' in files:
                base_build_dir = tmp_dir
            else:
                tmp_dir, subdir = os.path.split(tmp_dir)
                if subdir == '':
                    break

        if base_build_dir is not None:
            print('base_build_dir = ', base_build_dir)
            sys.path.insert(0, os.path.abspath(base_build_dir))

        from build_interface import InstallBUFRInterfaceECMWF
        # this will fail, because it loads the __init__.py inside
        # the pybufr_ecmwf directory, which in turn tries to load the
        # ecmwfbufr module, and that one does not yet exist, but will be
        # created by the ibi.build() call below.
        #from pybufr_ecmwf.build_interface import InstallBUFRInterfaceECMWF

        do_download_library_sources = False
        if str(self.download_library_sources).lower() == 'true':
            do_download_library_sources = True

        # run the build method from the InstallBUFRInterfaceECMWF class
        # defined in the custom build script, to build the extension module
        ibi = InstallBUFRInterfaceECMWF(
                  verbose=True,
                  preferred_fortran_compiler=self.preferred_fortran_compiler,
                  preferred_c_compiler=self.preferred_c_compiler,
                  fortran_compiler=self.fortran_compiler,
                  fortran_ld_library_path=self.fortran_ld_library_path,
                  fortran_flags=self.fortran_flags,
                  c_compiler=self.c_compiler,
                  c_ld_library_path=self.c_ld_library_path,
                  c_flags=self.c_flags,
                  debug_f2py_c_api=False,
                  download_library_sources=do_download_library_sources)

        # Build ecmwfbufr.so interface
        ibi.build()

        # remove all object files to prevent them from ending up
        # in the binary or rpm distribution packages
        ibi.clean()

        os.chdir(cwd)

        #print("self.distribution.dist_files = ", self.distribution.dist_files)
        #print("self.extensions[0].name = ", self.extensions[0].name)
        #print("dir(self.extensions[0]) = ", dir(self.extensions[0]))
        #sys.exit(1)

        #  #]

    #  #]

# modify the install_lib class to ensure the symlinks in the
# ecmwf_bufrtables dir are installed as symlinks and not copied as files
# (which would cause an excessive 1.5 GB of unneeded diskspace to be used)
class CustomInstallLib(_install_lib):
    #  #[ customised install_lib
    ''' a derived class that preserves symlinks when installing a
    python library '''
    def copy_tree(self, infile, outfile, preserve_mode=1, preserve_times=1,
                  preserve_symlinks=1, level=1):
        """ Run copy_tree with preserve_symlinks=1 as default """
        _install_lib.copy_tree(self, infile, outfile, preserve_mode,
                               preserve_times, preserve_symlinks, level)
    #  #]

#class Install(_install):
#    pass

# enable the disabled pylint warnings again
#pylint: enable-msg=W0201,R0902,R0904

DESCR = "a python interface around the ECMWF-BUFR library."
LONG_DESCR = """a python interface around the Fortran90 ECMWF-BUFR library
constructed using the f2py interface generation tool.
The equivalent subroutines to the ones in the ECMWF-BUFR
library are made available to python, but also a set of wrapper
routines/classes is implemented to
create also a more object-oriented/pythonic interface.
Building the interface is still a bit rough, and may require some
editing of the setup.cfg file to choose the correct fortran and c-compiler.
Alternatively some dedicated commandline options have been
added to the setup.py script to allow changing these settings.
Run 'setup.py build --help' to get a list of the currently available
build options.
"""

# define the list of classifiers
CL = ["Development Status :: 3 - Alpha",
      "Environment :: Console",
      "Intended Audience :: Science/Research",
      "License :: OSI Approved :: GNU General Public License (GPL)",
      "Natural Language :: English",
      "Operating System :: POSIX",
      "Programming Language :: Fortran",
      "Programming Language :: Python :: 2",
      "Programming Language :: Python :: 3",
      "Topic :: Scientific/Engineering",
      "Topic :: Software Development :: Libraries"
      ]

if sys.version_info[0] == 2:
    PACKAGE_NAME = 'pybufr-ecmwf'
elif sys.version_info[0] == 3:
    PACKAGE_NAME = 'pybufr-ecmwf3'
else:
    ERRTXT = 'This python version is not supported: '+sys.version
    raise NotImplementedError(ERRTXT)

# passing a python file to do the build does not work, it gives this error:
# error: unknown file type '.py' (from 'pybufr_ecmwf/build_interface.py')
ECMWF_BUFR_EXT = Extension('pybufr_ecmwf.ecmwfbufr',
                           ["pybufr_ecmwf/build_interface.py"])
#ECMWF_BUFR_EXT = Extension('pybufr_ecmwf.ecmwfbufr', [])

# make sure the alternative BUFR tables are included as well
ECMWF_BUFR_DATA = {'pybufr_ecmwf' : ['alt_bufr_tables/*.TXT',]}

setup(cmdclass={'build'       : Build,
                'build_ext'   : BuildExt,
                'install_lib' : CustomInstallLib},
      name=PACKAGE_NAME,
      version='0.80',
      description=DESCR,
      long_description=LONG_DESCR,
      author='Jos de Kloe',
      author_email='josdekloe@gmail.com',
      url='http://code.google.com/p/pybufr-ecmwf/',
      download_url="http://code.google.com/p/pybufr-ecmwf/source/checkout",
      classifiers=CL,
      platforms=["POSIX"],
      license="GPLv2",
      packages=['pybufr_ecmwf'],
      include_package_data=True,
      package_data=ECMWF_BUFR_DATA,
      ext_modules=[ECMWF_BUFR_EXT],
      requires=["numpy", "numpy.f2py", "subprocess"],
      provides=["pybufr_ecmwf"]
      # this requires use of the setup tools which needs to be installed
      # first (i.e. it makes the setup a little bit less portable)
      # see: http://peak.telecommunity.com/DevCenter/setuptools#test
      #test_suite = "pybufr_ecmwf.tests.run_unit_tests"
      )

#  #[ usage examples for setup.py:

# (see: http://docs.python.org/distutils/introduction.html)
# possible uses of this setup script:

# see: http://docs.python.org/distutils/sourcedist.html
# for more details on sdist commands.

# create a source distribution tar file: [works]
# ==>python setup.py sdist
# this creates a tarfile:  dist/pybufr-ecmwf-0.1.tar.gz
# and a MANIFEST file with a listing of all included files

# compile only the extension module pybufr_ecmwf/ecmwfbufr.so: [works]
# ==>python setup.py build_ext

# only recreate the MANIFEST file [works]
# python setup.py sdist --manifest-only

# see: http://docs.python.org/distutils/builtdist.html
# for more details on bdist commands.

# creation of a binary distribution tgz file: [works]
# ==>python setup.py bdist

# creation of an rpm file: [works!]
# ==>python setup.py bdist_rpm
# NOTE: for a locally installed python version the command should be:
# python2.6 ./setup.py bdist_rpm --python python2.6
# but this fails due to some path problem on my KNMI machine:
#  File not found: /nobackup/users/kloedej/temp_mercurial_repos/\
#     pybufr_ecmwf_copy/build/bdist.linux-i686/rpm/BUILD/pybufr-ecmwf-root/\
#     nobackup/users/kloedej/software_installed/scipy_numpy/lib/python2.6/\
#     site-packages/pybufr_ecmwf/ecmwfbufr.so

# build by an end user
# ==>python setup.py build

# installation by an end-user
# ==>python setup.py install

#  #]



#!/usr/bin/env python

"""
This is a small tool/example program intended to loop over all bufr messages
in a bufr file and extract the edition number from it, which is then printed
to stdout.
"""

# For details on the revision history, refer to the log-notes in
# the mercurial revisioning system hosted at google code.
#
# Written by: J. de Kloe, KNMI, Initial version 04-Feb-2014
#
# License: GPL v2.

#  #[ imported modules
import sys # operating system functions

# import the python file defining the RawBUFRFile class
from pybufr_ecmwf.raw_bufr_file import RawBUFRFile
from pybufr_ecmwf.bufr_interface_ecmwf import BUFRInterfaceECMWF

#  #]

def print_bufr_edition_number(input_bufr_file):
    #  #[
    """
    an example routine to demonstrate how to retrieve
    some meta data from the BUFR messages in a BUFR file
    """
    # get an instance of the RawBUFRFile class
    rbf = RawBUFRFile()

    # open the file for reading, count nr of BUFR messages in it
    # and store its content in memory, together with
    # an array of pointers to the start and end of each BUFR message
    rbf.open(input_bufr_file, 'rb')

    # extract the number of BUFR messages from the file
    num_msgs = rbf.get_num_bufr_msgs()

    for msg_nr in range(1, num_msgs+1):
        raw_msg = rbf.get_raw_bufr_msg(msg_nr)[0]
        bufr_obj = BUFRInterfaceECMWF(encoded_message=raw_msg)
        bufr_obj.decode_sections_012()
        bufr_edition = bufr_obj.ksec0[3-1]
        print 'BUFR msg %i has version %i' % (msg_nr, bufr_edition)

    # close the file
    rbf.close()
    #  #]

#  #[ run the tool
if len(sys.argv) < 2:
    print 'please give a BUFR file as argument'
    sys.exit(1)

INPUT_BUFR_FILE = sys.argv[1]

print_bufr_edition_number(INPUT_BUFR_FILE)
#  #]



#!/usr/bin/env python
"""
a small example program to demonstrate how a new set of BUFR tables
can be created from scratch, using the BufrTable class.
"""

from pybufr_ecmwf.bufr_table \
     import BufrTable, Descriptor, CompositeDescriptor, FlagDefinition

def create_bufr_tables():
    #  #[
    """
    a small function to hold the actual code to create a set of BUFR tables
    """

    # bt1 = BufrTable(autolink_tablesdir = "tmp_BUFR_TABLES")
    bt1 = BufrTable()

    # explanation on the reference number:
    # F=0 indicates this is a B-table entry
    # XX indicates the class or category within the table.
    # Private/local definitions should always be between 48 and 63
    # to avoid clashing with standard WMO descriptors
    # (see WMO_BUFR_Guide_Layer3-English-only section 3.1.2.1)
    # YYY should be in the range 000 to 255, but if a standard WMO class
    # is used it should be between 192 and 255 for local use.

    #            FXXYYY
    reference = '048001'
    name = 'my test variable A' # max. 64 characters
    unit = 'm/s' # SI units are recommended here. max.
    unit_scale = 3 # the power of 10 by which the element has to be
    #              # multiplied prior to encoding, or divided after
    #              # decoding.
    unit_reference = 1250 # the number to be subtracted from the element, after
    #                       scaling (if any) and prior to encoding (or added
    #                       after decoding)
    data_width = 4 # the number of bits used to store the data, which will
    #                always be converted to an integer (using the above
    #                scaling and reference subtraction)

    # NOTE: on linux the ECMWF BUFR library is always compiled using 32-bits
    # integers, so the max. data_width is 32 and the maximum unit_reference
    # is 2147483647 (2^31-1)

    descr_048001 = Descriptor(reference, name, unit, unit_scale,
                              unit_reference, data_width)
    bt1.add_to_B_table(descr_048001)

    #                 FXXYYY
    reference = '048002'
    name = 'my test variable B' # max. 64 characters
    unit = 'kg/m' # SI units are recommended here. max.
    unit_scale = 12
    unit_reference = 3456
    data_width = 5
    descr_048002 = Descriptor(reference, name, unit, unit_scale,
                              unit_reference, data_width)
    bt1.add_to_B_table(descr_048002)

    #            FXXYYY
    reference = '048003'
    name = 'a dummy counter' # max. 64 characters
    unit = 'FLAG TABLE 048003'
    unit_scale = 0
    unit_reference = 0
    data_width = 10
    descr_048003 = Descriptor(reference, name, unit, unit_scale,
                              unit_reference, data_width)
    bt1.add_to_B_table(descr_048003)

    #            FXXYYY
    reference = '048003'
    fldef = FlagDefinition(reference)
    fldef.flag_dict[0] = 'zero'
    fldef.flag_dict[1] = 'one'
    fldef.flag_dict[2] = 'two'
    fldef.flag_dict[3] = 'a very long line to describe the number 3; '*10
    fldef.flag_dict[4] = 'four'
    fldef.flag_dict[999] = 'missing'
    bt1.table_c[int(reference)] = fldef

    reference = '348001'
    descriptor_list = [descr_048001, descr_048002]
    comment = 'a small test D entry' # not written to file
    bufr_table_set = bt1
    descr_348001 = CompositeDescriptor(reference, descriptor_list,
                                       comment, bufr_table_set)

    bt1.add_to_D_table(descr_348001)

    print '='*50
    print "B-table:"
    print '='*50
    bt1.print_B_table()
    print '='*50
    print "C-table:"
    print '='*50
    bt1.print_C_table()
    print '='*50
    print "D-table:"
    print '='*50
    bt1.print_D_table()
    print '='*50

    # define the table name without preceding 'B' or 'D' character
    # (which will be prepended by the below write method)
    table_name = '_my_test_BUFR_table.txt'
    bt1.write_tables(table_name)
    #  #]

create_bufr_tables()
# end of this test program



#!/usr/bin/env python

"""
This file defines the RawBUFRFile class, an alternative for
loading and writing the binary raw BUFR messages from and to file
(needed because I cannot get the interface to the pb routines defined
in the ECMWF BUFR library to work in a portable way)
"""

#  #[ documentation
#
# Note about the use of the "#  #[" and "#  #]" comments:
#   these are folding marks for my favorite editor, emacs, combined with its
#   folding mode
#   (see http://www.emacswiki.org/emacs/FoldingMode for more details)
# Please do not remove them.
#
# For details on the revision history, refer to the log-notes in
# the mercurial revisioning system hosted at google code.
#
# Written by: J. de Kloe, KNMI (www.knmi.nl), Initial version 12-Nov-2009    
#
# License: GPL v2.
#
#  #]
#  #[ imported modules
from __future__ import (absolute_import, division,
                        print_function) #, unicode_literals)

import os          # operating system functions
import numpy as np # import numerical capabilities
import struct      # allow converting c datatypes and structs
from .helpers import python3
#  #]

class RawBUFRFile:
    #  #[
    """
    a class to read and write the binary BUFR messages from and
    to file. Is is intended to replace the pbio routines from the ECMWF
    library which for some obscure reason cannot be interfaced
    easily to python using the f2py tool.
    """
    def __init__(self, verbose = False,
                 warn_about_bufr_size = True):
        #  #[
        self.bufr_fd  = None
        self.filename = None
        self.filemode = None
        self.filesize = None
        self.data     = None
        self.list_of_bufr_pointers = []
        self.nr_of_bufr_messages = 0
        self.last_used_msg = 0
        self.verbose = verbose
        self.warn_about_bufr_size = warn_about_bufr_size
        #  #]
    def print_properties(self, prefix = "BUFRFile"):
        #  #[
        """
        print the properties of the current RawBUFRFile instance
        """
        # this one causes trouble with the unittesting since it gives
        # different addresses each time, and is not so very interesting
        # to print, so leave it out for now
        #print(prefix+": bufr_fd  = ", self.bufr_fd)
        print(prefix+": filename = {}".format(self.filename))
        print(prefix+": filemode = {}".format(self.filemode))
        print(prefix+": filesize = {}".format(self.filesize))
        if (self.data != None):
            print(prefix+": len(data) = {}".format(len(self.data)))
        else:
            print(prefix+": data = {}".format(self.data))
        print(prefix+": list_of_bufr_pointers = {}".
              format(self.list_of_bufr_pointers))
        print(prefix+": nr_of_bufr_messages = {}".
              format(self.nr_of_bufr_messages))
        #  #]
    def open(self, filename, mode, silent = False):
        #  #[
        """
        open a BUFR file to allow reading or writing raw BUFR messages
        """
        # note: the silent switch is only intended to suppress
        # warning and error messages during unit testing.
        # During normal use it should never be set to True.
        
        self.filename = filename
        self.filemode = mode
        
        # filename should include the path specification as well
        assert(mode in ['rb', 'wb', 'ab'])

        if (mode == 'rb'):
            if (os.path.exists(filename)):
                self.filesize = os.path.getsize(filename)
            else:
                if (not silent):
                    print("ERROR in BUFRFile.open():")
                    print("Opening file: ", self.filename, " with mode: ",
                          self.filemode, " failed")
                    print("This file was not found or is not accessible.")
                raise IOError
        elif (mode == 'wb'):
            self.filesize = 0
        elif (mode == 'ab'):
            # when appending it is allowed to have a non-existing
            # file, in which case one will be generated, so test for
            # this condition
            if (os.path.exists(filename)):
                # in this case, try to find out the amount of BUFR messages
                # already present in this file, by temporary opening
                # it in reading mode
                tmp_bf = RawBUFRFile()
                tmp_bf.open(filename, 'rb')
                #tmp_bf.print_properties(prefix = "tmp_bf (opened for reading)")
                count = tmp_bf.get_num_bufr_msgs()
                tmp_bf.close()
                del(tmp_bf)

                # then store the found number for later use
                self.nr_of_bufr_messages = count
                self.filesize = os.path.getsize(filename)

                if ((count == 0) and (self.filesize>0)):
                    print("WARNING: appending to non-zero file, but could")
                    print("not find any BUFR messages in it. Maybe you are")
                    print("appending to a non-BUFR file??")
            else:
                self.filesize = 0            

        try:
            self.bufr_fd = open(filename, mode)
        except:
            if (not silent):
                print("ERROR in BUFRFile.open():")
                print("Opening file: ", self.filename, " with mode: ",
                      self.filemode, " failed")
            raise IOError

        if (mode == 'rb'):
            try:
                self.data = self.bufr_fd.read()
            except:
                if (not silent):
                    print("ERROR in BUFRFile.open():")
                    print("Reading data from file: ", self.filename,
                          " with mode: ", self.filemode, " failed")
                raise IOError

            # split in separate BUFR messages
            self.split()

        #  #]
    def close(self):
        #  #[
        """
        close a BUFR file
        """
        # close the file
        self.bufr_fd.close()
        # then erase all settings
        self.__init__()
        #  #]
    def get_expected_msg_size(self, start_location):
        #  #[
        """ a routine to extract the expected message size from
        a BUFR message, needed to verify which start tag BUFR matches
        which end tag 7777. It also helps in excluding corrupted
        and falsely identified messages.
        """
        # According to ECMWF's BUFR User Guide:
        # http://www.ecmwf.int/products/data/software/bufr_user_guide.pdf
        # the first 8 bytes of a BUFR message should contain:
        # 1-4 : the four letters BUFR
        # 5-7 : total length of BUFR msg in bytes
        # 8   : BUFR Edition number (newest one currently is 4)
        #       valid edition numbers are 0,1,2,3,4
        # However, the oldest edition numbers 0 and 1 are exceptions
        # they do not have the full message lenght in section0, and in
        # fact only have a 4 byte section 0.
        # See the WMO BUFR Guide
        # http://www.wmo.int/pages/prog/www/WMOCodes/...
        #        Guides/BUFRCREX/Layer3-English-only.pdf
        # which is available from:
        # http://www.wmo.int/pages/prog/www/WMOCodes/...
        #        Guides/BUFRCREXPreface_en.html
        # Luckily the fourth byte of section 1 still gives the edition
        # number for these cases (although in later editions this byte
        # indicates the Bufr master Table version)
        # For these editions it is needed to retrieve the sections lenghts by
        # scanning each of the 6 sections and add them manually ...
        
        # Note: I hope to never see cray-blocked files again in my life
        # so this routine will not properly handle these ...
        # (these types of BUFR files do not conform to the BUFR
        #  standard anyway, so nobody should use them, but they might
        #  pop-up now and then from some old archive)
        # If you need to use this kind of files, write your own correction
        # routine to remove the 8 byte control-words inserted at every
        # 4096 bytes by this weird fileformat before trying to use them.

        # section nr               0 1 2 3 4 5
        section_sizes           = [0,0,0,0,0,0]
        section_start_locations = [0,0,0,0,0,0]

        # self.verbose = True
        if (self.verbose):
            print('getting size of BUFR message at start location: ',
                  start_location)
        try:
            raw_edition_number = self.data[start_location+8-1:start_location+8]
            edition_number = ord(raw_edition_number)
            if (self.verbose):
                print('edition_number = ', edition_number)
        except IndexError:
             # 0 signals this is not a valid BUFR msg, might be a false
             # start BUFR string, or a corrupted or truncated file
            return (0, section_sizes, section_start_locations)

        # note: the headers seem to use big-endian encoding
        # even on little endian machines, for the msg size.
        dataformat = ">1i"

        try:
            #  #[ retrieve size of section 0 and if possible the msg_size
            start_section0 = start_location
            if edition_number > 1:
                # get bytes 5 to 7 which should hold the total length of the
                # current BUFR message
                if python3:
                    raw_bytes = b'\0'+self.data[start_section0+5-1:
                                                start_section0+7]
                else:
                    raw_bytes = chr(0)+self.data[start_section0+5-1:
                                                 start_section0+7]
                try:
                    msg_size = struct.unpack(dataformat, raw_bytes)[0]
                except:
                    # 0 signals this is not a valid BUFR msg, might be a false
                    # start BUFR string, or a corrupted or truncated file
                    return (0, section_sizes, section_start_locations)

                #if self.verbose:
                #    print('section 0, byte 5: ', ord(raw_bytes[1]))
                #    print('section 0, byte 6: ', ord(raw_bytes[2]))
                #    print('section 0, byte 7: ', ord(raw_bytes[3]))
            if edition_number <= 1:
                size_section0 = 4
                # i.e. only the 4 characters 'BUFR'
            else:
                size_section0 = 8
                # i.e. the 4 characters 'BUFR',
                # AND the BUFR msg size (3 bytes) and the BUFR edition nr.

            if (self.verbose):
                print('size_section0 = ', size_section0,
                      'start_section0 = ', start_section0,
                      '[', hex(start_section0), ']')
            #  #]
            #  #[ retrieve size of section 1
            offset = size_section0
            start_section1 = start_location + offset
            # get length of section 1 from bytes 1 to 3
            raw_bytes = b'\x00'+self.data[start_section1+1-1:
                                          start_section1+3]
            try:
                size_section1 = struct.unpack(dataformat, raw_bytes)[0]
            except:
                # 0 signals this is not a valid BUFR msg, might be a false
                # start BUFR string, or a corrupted or truncated file
                return (0, section_sizes, section_start_locations)

            if (self.verbose):
                print('size_section1 = ', size_section1,
                      ' start_section1 = ', start_section1,
                      '[', hex(start_section1), ']')
            #  #]
            #  #[ retrieve size of section 2
            
            # see if the optional section 2 is present or not
            # this is indicated by bit 1 of byte 10 of section 1
            # bit 1 is the most significant bit, and corresponds
            # to the byte having an integer value of 128

            # NOTE: the documentation is inconsistent on this flag...
            # The WMO document WMO_BUFR_Guide_Layer3-English-only.pdf
            # states in the table on page L3-8 (the 10th page in the pdf)
            # that byte 8 signals presence of section 2.
            # The ECMWF document bufr_user_guide.pdf states in the
            # table on page 7 that byte 10 signals the presence of section 2,
            # however, in the text below it states that byte 8 should be used!
            #
            # Looking at the actual fortran code in the ECMWF BUFR library
            # the situation seems more subtle, and depends on the
            # BUFR edition version used...
            # 
            # Edition      0 1 2 3 4       available in ksec0(3) and ksec1(2)
            # sec1 length  3 3 3 3 3 bytes copied to ksec1(1)
            # master table 1 1 1 1 1 bytes copied to ksec1(2) if edition<1
            #                              else it is copied to ksec1(14)
            # orig. table  2 2 2 2 4 bytes copied to ksec1(3) and ksec1(16) 
            # seq. number  1 1 1 1 1 bytes copied to ksec1(4)
            # sec2presence 1 1 1 1 1 bytes copied to ksec1(5)
            
            if edition_number < 4:
                byte_to_use = 8
            else:
                byte_to_use = 10

            sec2_presence_flag = \
                 ord(self.data[start_location+offset+byte_to_use-1:
                               start_location+offset+byte_to_use])
                               
            section2_present = False

            #if (sec2_presence_flag & 1):
            if (sec2_presence_flag > 0):
                section2_present = True
                
            # retrieve size of section 2
            offset = size_section0 + size_section1
            start_section2 = start_location + offset
            if section2_present:
                # get length of section 2 from bytes 1 to 3
                raw_bytes = chr(0)+self.data[start_section2+1-1:
                                             start_section2+3]
                try:
                    size_section2 = struct.unpack(dataformat, raw_bytes)[0]
                except:
                    # 0 signals this is not a valid BUFR msg, might be a false
                    # start BUFR string, or a corrupted or truncated file
                    return (0, section_sizes, section_start_locations)
            else:
                size_section2 = 0

            if (self.verbose):
                print('size_section2 = ',size_section2,
                      ' start_section2 = ', start_section2,
                      '[', hex(start_section2), ']')
            #  #]
            #  #[ retrieve size of section 3
            offset = size_section0 + size_section1 + size_section2
            start_section3 = start_location + offset
            # get length of section 3 from bytes 1 to 3
            raw_bytes = b'\x00'+self.data[start_section3+1-1:
                                          start_section3+3]
            try:
                size_section3 = struct.unpack(dataformat, raw_bytes)[0]
            except:
                # 0 signals this is not a valid BUFR msg, might be a false
                # start BUFR string, or a corrupted or truncated file
                return (0, section_sizes, section_start_locations)

            if (self.verbose):
                print('size_section3 = ',size_section3,
                      ' start_section3 = ', start_section3,
                      '[', hex(start_section3), ']')
            #  #]
            #  #[ retrieve size of section 4
            offset = size_section0 + size_section1 + \
                     size_section2 + size_section3
            start_section4 = start_location + offset
            # get length of section 4 from bytes 1 to 3
            raw_bytes = b'\x00'+self.data[start_section4+1-1:
                                          start_section4+3]
            #if self.verbose:
            #    print('section 4, byte 1: ',
            #          ord(raw_bytes[1]), hex(ord(raw_bytes[1])))
            #    print('section 4, byte 2: ',
            #          ord(raw_bytes[2]), hex(ord(raw_bytes[1])))
            #    print('section 4, byte 3: ',
            #          ord(raw_bytes[3]), hex(ord(raw_bytes[1])))

            try:
                size_section4 = struct.unpack(dataformat, raw_bytes)[0]
            except:
                # 0 signals this is not a valid BUFR msg, might be a false
                # start BUFR string, or a corrupted or truncated file
                return (0, section_sizes, section_start_locations)

            if (self.verbose):
                print('size_section4 = ',size_section4,
                      ' start_section4 = ', start_section4,
                      '[', hex(start_section4), ']')
            #  #]
            #  #[ retrieve size of section 5
            offset = size_section0 + size_section1 + \
                     size_section2 + size_section3 + size_section4
            start_section5 = start_location + offset

            size_section5 = 4

            if (self.verbose):
                print('size_section5 = ',size_section5,
                      ' start_section5 = ', start_section5)
            #  #]
            calculated_msg_size = size_section0 + size_section1 + \
                                  size_section2 + size_section3 + \
                                  size_section4 + size_section5

            if edition_number <= 1:
                # for editions 0 and 1 the msg_size is not contained
                # in section 0, and still needs to be set:
                msg_size = calculated_msg_size
            else:
                # extra sanity check
                if msg_size != calculated_msg_size:
                    print('ERROR! msg_size from section0 does not match')
                    print('msg_size calculated from individual '+
                          'section lengths!')
                    print('msg_size from section0: ',msg_size)
                    print('calculated msg_size;    ',calculated_msg_size)
                    print('SKIPPING this message...')
                    return (0, section_sizes, section_start_locations)
                
        except IndexError:
            # 0 signals this is not a valid BUFR msg, might be a false
            # start BUFR string, or a corrupted or truncated file
            return (0, section_sizes, section_start_locations)
        
        section_sizes           = [size_section0, size_section1,
                                   size_section2, size_section3,
                                   size_section4, size_section5]
        # note: the start locations stored above are relative to
        # the beginning of the file, but for further use it is more
        # convenient to have them relative to the beginning
        # of the BUFR message, so subtract start_location  from them
        section_start_locations = [start_section0-start_location,
                                   start_section1-start_location,
                                   start_section2-start_location,
                                   start_section3-start_location,
                                   start_section4-start_location,
                                   start_section5-start_location]
        
        # see: p.L3-3 (p.5) of the file Layer3-English-only.pdf
        #      mentioned above.
        # 
        # this limit was raised from 15kb to 500kb on 7-Nov-2007
        # see: http://www.wmo.int/pages/prog/www/ois/Operational_Information/ \
        #             Newsletters/2000_2009/2005/Sept05/GTS.html

        if self.warn_about_bufr_size:
            if msg_size > 500000:
                print("WARNING: by convention BUFR messages should not be "+
                      "larger than 500kb to allow transmission over the GTS. "+
                      "Size of current message is: ", msg_size, " bytes")

        return (msg_size, section_sizes, section_start_locations)
        #  #]        
    def split(self):
        #  #[
        """
        scan a BUFR file to detect the start and end locations of the
        separate BUFR messages. Note that a BUFR file may contain
        additional junk, like GTS headers and such. The code should be
        robust enough to handle this.
        """
        # Purpose: scans the file for the string "BUFR"
        # which indicate the start of a new BUFR message,
        # counts the nr of BUFR messages, and stores file
        # pointers to the start of each BUFR message.

        # safety catch
        if (self.filesize == 0):
            self.nr_of_bufr_messages = 0
            return

        # note: this very simpple search algorithm might accidently
        # find the string "7777" in the middle of the data of a BUFR message.
        # To check on this, make sure the distance between the end of a
        # message and the start of a message if either 0 or 2 bytes
        # (this may happen if the file is padded with zeros to contain
        #  a multiple of 4 bytes)
        # Do the same check on the end of the file.

        txt_start  = b'BUFR'
        txt_end    = b'7777'
        list_of_start_locations = []
        list_of_end_locations   = []

        # try to find the start strings
        search_pos = 0
        file_end_reached = False
        while not file_end_reached:
            start_pos = self.data.find(txt_start, search_pos)
            if (start_pos == -1):
                file_end_reached = True
            else:
                list_of_start_locations.append(start_pos)
                search_pos = start_pos + 4

        # try to find the end strings
        search_pos = 0
        file_end_reached = False
        while not file_end_reached:
            end_pos = self.data.find(txt_end, search_pos)
            if (end_pos == -1):
                file_end_reached = True
            else:
                list_of_end_locations.append(end_pos)
                search_pos = end_pos + 4

        self.list_of_bufr_pointers = []

        # try each BUFR message; extract its length and see if
        # it matches an end location. If not we found a false start
        # marker or a corrupt BUFR message. If it matches, assume
        # the BUFR message is valid and add it to the list
        for start_location in list_of_start_locations:
            expected_msg_size, section_sizes, section_start_locations = \
                               self.get_expected_msg_size(start_location)
            if (self.verbose):
                print('expected_msg_size = ', expected_msg_size)
            expected_msg_end_location = start_location + expected_msg_size - 4
            if expected_msg_end_location in list_of_end_locations:
                if (self.verbose):
                    print('message seems alright, adding it to the list')
                # point to the end of the four sevens
                # (in slice notation, so the bufr msg data
                # can be adressed as data[start_pos:end_pos])
                # and store it
                self.list_of_bufr_pointers.append((start_location,
                                                   expected_msg_end_location+4,
                                                   section_sizes,
                                                   section_start_locations))
        

        # count howmany we found
        self.nr_of_bufr_messages = len(self.list_of_bufr_pointers)

        if (self.verbose):
            print("list_of_start_locations = ", list_of_start_locations)
            print("list_of_end_locations   = ", list_of_end_locations)

        #  #]
    def get_num_bufr_msgs(self):
        #  #[
        """
        request the number of BUFR messages in the current file
        """
        if (self.bufr_fd == None):
            print("ERROR: a bufr file first needs to be opened")
            print("using BUFRFile.open() before you can request the")
            print("number of BUFR messages in a file ..")
            raise IOError

        return self.nr_of_bufr_messages
        #  #]
    def get_raw_bufr_msg(self, msg_nr):
        #  #[
        """
        get the raw data for the BUFR message with given msg_nr
        (start counting at 1)
        """
        
        if (self.bufr_fd == None):
            print("ERROR: a bufr file first needs to be opened")
            print("using BUFRFile.open() before you can use the raw data ..")
            raise IOError

        # sanity test
        if (msg_nr>self.nr_of_bufr_messages):
            print("WARNING: non-existing BUFR message: ", msg_nr)
            print("This file only contains: ", self.nr_of_bufr_messages,
                  " BUFR messages")
            return (None, None, None)

        if (msg_nr<1):
            print("WARNING: invalid BUFR message number: ", msg_nr)
            print("For this file this number should be between 1 and: ",
                  self.nr_of_bufr_messages)
            return (None, None, None)

        self.last_used_msg = msg_nr
        (start_index, end_index, section_sizes, section_start_locations) = \
                      self.list_of_bufr_pointers[msg_nr-1]

        size_bytes = (end_index-start_index)

        # +3 because we have to round upwards to make sure all
        # bytes fit into the array of words (otherwise the last
        # few might be truncated from the data, which will crash
        # the struct.unpack() call below)
        size_words = (size_bytes+3)//4
        padding_bytes = size_words*4-size_bytes

        if (self.verbose):
            print("size_bytes = ", size_bytes)
            print("size_words = ", size_words)
            print("size_words*4 = ", size_words*4)
            print("padding_bytes = ", padding_bytes)
            
        # make sure we take the padding bytes along
        end_index = end_index+padding_bytes
        
        raw_data_bytes = self.data[start_index:end_index]

        # make sure the raw datastream is padded with zero bytes
        # to a multiple of 4 bytes. The ECMWF software may crash
        # if this is not the case ...
        nbytes = len(raw_data_bytes)
        nbytes_rounded = 4*(nbytes//4)
        if nbytes != nbytes_rounded:
            # print('padding problem found! fixing it ...')
            # print('nbytes = ',nbytes)
            # print('nbytes_rounded = ',nbytes_rounded)
            num_zeros_to_add = nbytes_rounded+4-nbytes
            # print('num_zeros_to_add = ',num_zeros_to_add)
            if python3:
                str_to_add = b''.join(b'\x00' for i in range(num_zeros_to_add))
            else:
                str_to_add = ''.join(b'\x00' for i in range(num_zeros_to_add))

            raw_data_bytes = raw_data_bytes + str_to_add

        if (self.verbose):
            print("len(raw_data_bytes) = ", len(raw_data_bytes))

        # assume little endian for now when converting
        # raw bytes/characters to integers and vice-versa
        dataformat = "<"+str(size_words)+"i"
        try:
            words = np.array(struct.unpack(dataformat, raw_data_bytes))
        except:
            # 0 signals this is not a valid BUFR msg, might be a false
            # start BUFR string, or a corrupted or truncated file
            return (0, section_sizes, section_start_locations)

        return (words, section_sizes, section_start_locations)
        #  #]
    def get_next_raw_bufr_msg(self):
        #  #[
        """
        get the raw data for the next BUFR message.
        This routine uses the internal instance variable last_used_msg
        to store the index of the last read BUFR message.
        """
        if (self.last_used_msg == self.nr_of_bufr_messages):
            raise EOFError
        
        return self.get_raw_bufr_msg(self.last_used_msg+1)
        #  #]
    def write_raw_bufr_msg(self, words):
        #  #[
        """
        write the raw BUFR message data to the BUFR file
        """
        # input data should be an array of words!
        size_words = len(words)
        size_bytes = size_words*4
        if (self.verbose):
            print("size_bytes = ", size_bytes)
            print("size_words = ", size_words)

        # convert the words to bytes in a string and write them to file

        # question: is this conversion really needed, or could I also just
        # directly write the data as words?
        # Answer: yes this really is needed! If the words are just written
        # as such, python converts them to long integers and writes
        # 8 bytes for each word in stead of 4 !!!!!
        # activate the next 2 lines to test this:
        #self.bufr_fd.write(words)
        #return
        
        # assume little endian for now when converting
        # raw bytes/characters to integers and vice-versa
        dataformat = "<i"
        for (i, word) in enumerate(words):
            data = struct.pack(dataformat, word)
            self.bufr_fd.write(data)

            if i == 0:
                if (self.verbose):
                    print("word = ", word)
                    print('data = ', data)
                    print('data[:4] = ', data[:4])
                    print('data[:4] = ', ';'.join(str(data[j])
                                                  for j in range(4)
                                                  if data[j].isalnum()))
                # safety check
                assert(data[:4] == b'BUFR')

        self.nr_of_bufr_messages = self.nr_of_bufr_messages + 1
        self.filesize = self.filesize + size_bytes
        #  #]
    #  #]



#!/usr/bin/env python

"""
this is a stripped down version of the f2py tool that is present on
my own machine, and is included here since it seems the f2py tool
itself is not available in the standard path on some machines,
even if numpy and numpy.f2py are available as python module.
"""

from numpy.f2py import main
main()



#!/usr/bin/env python

"""
This is a small tool/example program intended to loop over all bufr messages
in a bufr file and extract all the data from it, which is then printed
to stdout or written to file, either in ascii or csv format.
"""

# For details on the revision history, refer to the log-notes in
# the mercurial revisioning system hosted at google code.
#
# Written by: J. de Kloe, KNMI, Initial version 24-Sep-2010
#
# License: GPL v2.

#  #[ imported modules
import sys # operating system functions
import getopt # a simpler version of argparse, which was introduced in
              # python 2.7, and is not by default available for older versions

# import the python file defining the RawBUFRFile class
from pybufr_ecmwf.bufr import BUFRReader
from pybufr_ecmwf.raw_bufr_file import RawBUFRFile
from pybufr_ecmwf.bufr_interface_ecmwf import BUFRInterfaceECMWF
from pybufr_ecmwf.helpers import python3

#  #]

def print_bufr_content1(input_bufr_file, output_fd, separator, max_msg_nr):
    #  #[ implementation 1
    """
    example implementation using the BUFRReader class
    combined with the get_values_as_2d_array method
    """

    # get an instance of the BUFR class
    # which automatically opens the file for reading and decodes it
    bob = BUFRReader(input_bufr_file, warn_about_bufr_size=False)

    msg_nr = 0
    while True:
        try:
            bob.get_next_msg()
            msg_nr += 1
        except EOFError:
            break

        # add header strings
        # print 'DEBUG: bob.msg_loaded ',bob.msg_loaded
        list_of_names = []
        list_of_units = []
        list_of_names.extend(bob.get_names())
        list_of_units.extend(bob.get_units())
        list_of_unexp_descr = bob.bufr_obj.py_unexp_descr_list

        if bob.msg_loaded == 1:
            output_fd.write(separator.join(list_of_names) + "\n")
            output_fd.write(separator.join(list_of_units) + "\n")
            list_of_unexp_descr_first_msg = bob.bufr_obj.py_unexp_descr_list

        data = bob.get_values_as_2d_array()

        if list_of_unexp_descr != list_of_unexp_descr_first_msg:
            print '\n\n'
            print 'WARNING: it seems different types of BUFR messages'
            print 'are mixed in this BUFR file, meaning that the list of'
            print 'descriptor names and units printed on the first 2 output'
            print 'lines will not match with all lines of data.'
            print 'To prevent confusion, therefore decoding is halted'
            print 'It is recommended to first sort BUFR messages by type'
            print 'before converting them to ascii or csv.'
            print 'The example script soft_bufr_msgs.py can be used'
            print 'sort a BUFR file.'
            print '\n\n'
            print 'Detailed info:'
            print 'list_of_unexp_descr != list_of_unexp_descr_first_msg !'
            print 'list_of_unexp_descr           = ', \
                  list_of_unexp_descr
            print 'list_of_unexp_descr_first_msg = ', \
                  list_of_unexp_descr_first_msg
            sys.exit(0)

        if data.shape[0]*data.shape[1] == 0:
            print 'NO DATA FOUND! this seems an empty BUFR message !'
            continue

        for subs in range(len(data[:, 0])):
            output_fd.write(str(subs)+separator+
                            separator.join(str(val) for val in data[subs, :])+
                            "\n")
        print 'converted BUFR msg nr. ', msg_nr
        if (max_msg_nr > 0) and (msg_nr >= max_msg_nr):
            print 'skipping remainder of this BUFR file'
            break

    # close the file
    bob.close()
    if msg_nr == 0:
        print 'no BUFR messages found, are you sure this is a BUFR file?'
    #  #]

def print_bufr_content2(input_bufr_file, output_fd, separator, max_msg_nr):
    #  #[ implementation 2
    """
    example implementation using the BUFRReader class
    combined with the get_value method
    """

    # get an instance of the BUFR class
    # which automatically opens the file for reading and decodes it
    bob = BUFRReader(input_bufr_file, expand_flags=True)
    msg_nr = 0
    while True:
        try:
            bob.get_next_msg()
            msg_nr += 1
        except EOFError:
            break

        # add header strings
        if bob.msg_loaded == 1:
            list_of_names = []
            list_of_units = []
            list_of_names.extend(bob.get_names())
            list_of_units.extend(bob.get_units())
            output_fd.write(separator.join(list_of_names) + "\n")
            output_fd.write(separator.join(list_of_units) + "\n")

        nsubsets = bob.get_num_subsets()
        for subs in range(nsubsets):
            nelements = bob.get_num_elements()
            data_list = []
            for descr_nr in range(nelements):
                data = bob.get_value(descr_nr, subs)
                data_list.append(data)
            output_fd.write(str(subs)+separator+
                            separator.join(str(val) for val in data_list)+
                            "\n")

        print 'converted BUFR msg nr. ', msg_nr
        if (max_msg_nr > 0) and (msg_nr >= max_msg_nr):
            print 'skipping remainder of this BUFR file'
            break

    # close the file
    bob.close()
    if msg_nr == 0:
        print 'no BUFR messages found, are you sure this is a BUFR file?'
    #  #]

def print_bufr_content3(input_bufr_file, output_fd, separator, max_msg_nr):
    #  #[ implementation 3
    """
    example implementation using the BUFRInterfaceECMWF class
    """

    # get an instance of the RawBUFRFile class
    rbf = RawBUFRFile()

    # open the file for reading, count nr of BUFR messages in it
    # and store its content in memory, together with
    # an array of pointers to the start and end of each BUFR message
    rbf.open(input_bufr_file, 'rb')

    # extract the number of BUFR messages from the file
    num_msgs = rbf.get_num_bufr_msgs()

    # print 'num_msgs = ',num_msgs

    for msg_nr in range(1, num_msgs+1):
        encoded_message, section_sizes, section_start_locations = \
                         rbf.get_raw_bufr_msg(msg_nr)
        bufr_obj = BUFRInterfaceECMWF(encoded_message, section_sizes,
                                      section_start_locations)
        #                              verbose=True)
        bufr_obj.decode_sections_012()
        bufr_obj.setup_tables()
        # print 'num_subsets: ', bufr_obj.get_num_subsets()
        # print 'num_elements: ',bufr_obj.get_num_elements()
        # bufr_obj.decode_sections_0123()
        # bufr_obj.print_sections_0123_metadata()

        # d = '/home/jos/werk/pybufr_ecmwf_interface/'+\
        #     'BUFR_test_files/radar/bufrtables/'
        # bufr_obj.setup_tables(table_b_to_use = d+'B0000000000085011012.TXT',
         #                      table_d_to_use = d+'D0000000000085011012.TXT')
        # bufr_obj.print_sections_012()
        # bufr_obj.fill_descriptor_list()
        bufr_obj.decode_data()

        # Create header lines from variable names and units
        if msg_nr == 1:
            list_of_names = []
            list_of_units = []
            for (cname, cunit) in zip(bufr_obj.cnames, bufr_obj.cunits):
                # glue the ndarray of characters together to form strings
                if python3:
                    cname_str = ''.join(c.decode() for c in cname).strip()
                    cunit_str = ''.join(c.decode() for c in cunit).strip()
                else:
                    cname_str = ''.join(cname).strip()
                    cunit_str = ''.join(cunit).strip()

                # append the strings to the head list and quote them
                list_of_names.append('"'+cname_str+'"')
                list_of_units.append('"'+cunit_str+'"')

            output_fd.write(separator.join(list_of_names) + '\n')
            output_fd.write(separator.join(list_of_units) + '\n')

        nsubsets = bufr_obj.get_num_subsets()
        for subs in range(nsubsets):
            nelements = bufr_obj.get_num_elements()
            data_list = []
            for descr_nr in range(nelements):
                data = bufr_obj.get_value(descr_nr, subs)
                data_list.append(data)
            output_fd.write(str(subs)+separator+
                            separator.join(str(val) for val in data_list)+
                            "\n")
        print 'converted BUFR msg nr. ', msg_nr
        if (max_msg_nr > 0) and (msg_nr >= max_msg_nr):
            print 'skipping remainder of this BUFR file'
            break

    # close the BUFR file
    rbf.close()
    if num_msgs == 0:
        print 'no BUFR messages found, are you sure this is a BUFR file?'

    #  #]

def usage():
    #  #[
    """ a small routine to print the options that may be used
    with this example progra,
    """
    print 'Usage: '
    print sys.argv[0] + ' [OPTIONS]'
    print ''
    print 'With [OPTIONS] being one or more of these possibilities: '
    print '-a or --ascii    selects ascii output'
    print '-c or --csv      selects csv output'
    print '-i or --infile   defines the input BUFR file to be used [required]'
    print '-o or --outfile  defines the output file to be used'
    print '                 if this option is omitted, stdout will be used'
    print '-1 or -2 or -3   test implementation 1, 2 or 3 [default is 1]'
    print '-m or --maxmsgnr defines max number of BUFR messages to convert'
    print '-h               display this help text'
    #  #]

def main():
    #  #[ define the main program
    """ define the main code for this test program as a subroutine
    to prevent the ugly pylint convention of using capital letters
    for all variables at root level.
    """
    try:
        # command line handling; the ':' and '=' note that the
        # options must have a value following it
        short_options = 'aci:o:m:h123'
        long_options = ['ascii', 'csv', 'infile=', 'outfile=',
                        'maxmsgnr=', 'help']
        (options, other_args) = getopt.getopt(sys.argv[1:],
                                              short_options, long_options)
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit(2)

    # test prints
    #print 'options = ', options
    #print 'other_args = ', other_args

    # defaults
    output_to_ascii = True
    input_bufr_file = None
    output_file = None
    implementation_nr = 1
    max_msg_nr = -1

    for (opt, value) in options:
        if   (opt == '-h') or (opt == '--help'):
            usage()
        elif (opt == '-a') or (opt == '--ascii'):
            output_to_ascii = True
        elif (opt == '-c') or (opt == '--csv'):
            output_to_ascii = False # implies csv
        elif (opt == '-i') or (opt == '--infile'):
            input_bufr_file = value
        elif (opt == '-o') or (opt == '--outfile'):
            output_file = value
        elif opt == '-1':
            implementation_nr = 1
        elif opt == '-2':
            implementation_nr = 2
        elif opt == '-3':
            implementation_nr = 3
        elif (opt == '-m') or (opt == '--maxmsgnr'):
            max_msg_nr = int(value)
        else:
            print "Unhandled option: "+opt
            usage()
            sys.exit(2)

    # ensure input_bufr_file is defined
    if input_bufr_file is None:
        print "Missing input file!"
        usage()
        sys.exit(2)

    # warn about unused arguments
    if len(other_args) > 0:
        print 'WARNING: there seem to be unused arguments:'
        print other_args

    # Open the output file
    if output_file:
        output_fd = open(output_file, "w")
    else:
        output_fd = sys.stdout

    if output_to_ascii:
        separator = ' ' # ascii case
    else:
        separator = ',' # csv case

    if implementation_nr == 1:
        print_bufr_content1(input_bufr_file, output_fd,
                            separator, max_msg_nr)
    elif implementation_nr == 2:
        print_bufr_content2(input_bufr_file, output_fd,
                            separator, max_msg_nr)
    elif implementation_nr == 3:
        print_bufr_content3(input_bufr_file, output_fd,
                            separator, max_msg_nr)

    if output_file:
        # close the output file
        output_fd.close()

        if output_to_ascii:
            print "ascii output written to file " + output_file
        else:
            print "csv output written to file " + output_file

#  #]

# run the tool
main()



#!/usr/bin/env python

"""
a little script to use the 2to3 conversion tool on this
source code, to allow easier porting to python3

WARNING:
numpy for python3 is still very new, so may not be available
as default package from your favourite linux distribution for some time.
Therefore you need to install a 3rd party numpy rpm/deb package or install
numpy from source first, before you can use the python3 version of this
pybufr_ecmwf module. JK, 22-Apr-2011.
"""
from __future__ import (absolute_import, division,
                        print_function) # , unicode_literals)
import os, sys
import subprocess  # support running additional executables

PY3_CONVERTED_PATH = 'tmp_2to3_converted_sources'

def run_shell_command(cmd, libpath=None,
                      module_path='./', verbose=True):
    #  #[
    """ a wrapper routine around subprocess.Popen intended
    to make it a bit easier to call this functionality.
    Options:
    -libpath: add this path to the LD_LIBRARY_PATH environment variable
     before executing the subprocess
    -catch_output: if True, this function returns 2 lists of text lines
     containing the stdout and stderr of the executed subprocess
    -verbose: give some feedback to the user while executing the
     code (usefull for debugging)"""

    # get the list of already defined env settings
    env = os.environ
    if libpath:
        # add the additional env setting
        envname = "LD_LIBRARY_PATH"
        if envname in env:
            env[envname] = env[envname] + ":" + libpath
        else:
            env[envname] = libpath

    if 'PYTHONPATH' in env:
        env['PYTHONPATH'] = env['PYTHONPATH']+':'+module_path
    else:
        env['PYTHONPATH'] = module_path

    if verbose:
        print("Executing command: ", cmd)

    # print('env[PYTHONPATH] = ',env['PYTHONPATH'])
    subpr = subprocess.Popen(cmd,
                             shell=True,
                             env=env,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)

    # wait until the child process is done
    # subpr.wait() # seems not necessary when catching stdout and stderr

    lines_stdout = subpr.stdout.readlines()
    lines_stderr = subpr.stderr.readlines()

    #print("lines_stdout: ", lines_stdout)
    #print("lines_stderr: ", lines_stderr)

    return (lines_stdout, lines_stderr)
    #  #]

def port_2to3():
    #  #[
    '''
    a routine try to automatically convert the whole source code
    of this module to python3
    '''
    # first see if python3 and the 2to3 tool are available
    tools_to_check = ['2to3', 'python3', 'hg']

    # note: normally, if you have a recent python2 installed
    # (and on some systems maybe the python-tools package as well)
    # the 2to3 tool will be available in: /usr/bin/2to3
    #
    # However, if you have only python3 installed then (on my fedora14
    # system) 2to3 is installed in: /usr/lib64/python3.1/Tools/scripts/2to3
    # In my case 2to3 is part of the python3-tools package.

    for tool_to_check in tools_to_check:
        cmd = 'which '+tool_to_check
        (lines_stdout, lines_stderr) = run_shell_command(cmd, verbose=False)

        if len(lines_stderr) > 0:
            print('sorry, the '+tool_to_check+' tool seems not installed')
            print('num lines stderr/stdout = ', \
                  len(lines_stderr), len(lines_stdout))
            sys.exit(1)
        else:
            print('tool found: ', tool_to_check)

    # create the testdir
    if os.path.exists(PY3_CONVERTED_PATH):
        print('ERROR: testdir: ', PY3_CONVERTED_PATH, ' already exists')
        print('please first run the clean.py tool before trying to run')
        print('this conversion script')
        sys.exit(1)
    os.mkdir(PY3_CONVERTED_PATH)

    # clone the repository to the testdir
    print('cloning the repository')
    cmd = 'hg clone . '+PY3_CONVERTED_PATH
    (lines_stdout, lines_stderr) = run_shell_command(cmd, verbose=False)
    if len(lines_stderr) > 0:
        print('sorry, failed command: ', cmd)
        for line in lines_stderr:
            print(line, end='')
        sys.exit(1)

    # do the actual conversion
    print('converting sources to python3')
    cmd = '2to3 -w '+PY3_CONVERTED_PATH
    (lines_stdout, lines_stderr) = run_shell_command(cmd, verbose=False)
    # this next check is not usefull, since the 2to3 tool issues
    # several messages to stderr even if all runs well
    # if len(lines_stderr)>0:
    #    print('sorry, failed command: ', cmd)
    #    for line in lines_stderr:
    #        print(line, end='')
    #    sys.exit(1)

    # walk along all python files and fix the shebang/hashbang line
    # since the 2to3 tool seems to not fix this one.
    for walk_result in os.walk(PY3_CONVERTED_PATH):
        dirpath = walk_result[0]
        filenames = walk_result[2]
        for filename in filenames:
            ext = os.path.splitext(filename)[1]
            if ext == '.py':
                path_and_file = os.path.join(dirpath, filename)
                print('fixing: ', path_and_file)

                fdb = open(path_and_file) #, 'rt')
                lines = fdb.readlines()
                fdb.close()

                fda = open(path_and_file, 'w')
                for i, line in enumerate(lines):
                    if ('#!' in line) and (i == 0):
                        print('shebang line found: ', line.replace('\n', ''))
                        fda.write('#!/usr/bin/env python3\n')
                    else:
                        fda.write(line)
                fda.close()

    # commit the modified code to allow usage by my automatic test
    # system, which copies the module by taking a clone of the repository.
    cmd = 'cd '+PY3_CONVERTED_PATH+';'+\
          'hg commit -m "automatic commit by the port_2to3.py tool"'
    (lines_stdout, lines_stderr) = run_shell_command(cmd, verbose=False)

    print('conversion done')
    #  #]

# run the conversion tool
port_2to3()

# todo: try to run the build and test stages
